#include <cmath>
#include <cfloat>

#include "client/rendering/Engine.h"
#include "client/rendering/wrapper/GLFWWrapper.h"
#include "client/rendering/wrapper/GLWrapper.h"
#include "client/rendering/Renderer.h"
#include "client/math/Plane.h"

Engine::Engine(Shaders& shaders, Framebuffers& fb, Camera& camera, const WindowSize& size,
        RenderSettings& renderSettings) :
shaders(shaders), fb(fb), camera(camera), size(size), renderSettings(renderSettings), frustum(60.0f, 0.1f, 80.0f),
ssaoNoise(4, 4) {
    rectangle.add( {-1, -1, 0, 0, 0, 0, 0, 0});
    rectangle.add( {1, 1, 0, 1, 1, 0, 0, 0});
    rectangle.add( {-1, 1, 0, 0, 1, 0, 0, 0});
    rectangle.add( {-1, -1, 0, 0, 0, 0, 0, 0});
    rectangle.add( {1, -1, 0, 1, 0, 0, 0, 0});
    rectangle.add( {1, 1, 0, 1, 1, 0, 0, 0});
    rectangle.build();
}

void Engine::renderTick(float lag, const Game& game) {
    camera.update(lag);
    updateWorldProjection();
    updateWorldView();

    if(renderSettings.shadows) {
        renderShadow(lag, game);
    }
    renderWorld(lag, game);
    if(renderSettings.ssao) {
        renderSSAO();
    }
    renderPostWorld();
    renderTextOverlay(lag, game);
}

void Engine::renderShadow(float lag, const Game& game) {
    fb.shadow.bind();
    GLWrapper::enableDepthTesting();
    shaders.shadow.use();
    worldShadowProjView.set(worldShadowProj).mul(worldShadowView);
    shaders.shadow.setMatrix("projView", worldShadowProjView.getValues());
    Renderer renderer(shaders.shadow, model);
    game.renderWorld(lag, renderer);
}

void Engine::renderWorld(float lag, const Game& game) {
    fb.world.bind();
    GLWrapper::enableDepthTesting();
    shaders.world.use();

    Matrix rWorldShadowProjView;
    rWorldShadowProjView.translate(0.5f, 0.5f, 0.5f).scale(0.5f).mul(worldShadowProjView);

    shaders.world.setMatrix("projViewShadow", rWorldShadowProjView.getValues());
    shaders.world.setMatrix("proj", worldProj.getValues());
    shaders.world.setMatrix("view", worldView.getValues());
    model.clear();
    shaders.world.setMatrix("model", model.get().getValues());
    fb.shadow.bindDepthTexture(1);
    shaders.world.setInt("shadows", renderSettings.shadows);
    shaders.world.setFloat("radius", renderSettings.testRadius);
    shaders.world.setFloat("bias", renderSettings.testBias);
    Renderer renderer(shaders.world, model);
    game.renderWorld(lag, renderer);
}

void Engine::renderSSAO() {
    shaders.ssao.use();

    Matrix rProj;
    rProj.translate(0.5f, 0.5f, 0.5f).scale(0.5f).mul(worldProj);

    shaders.ssao.setMatrix("proj", rProj.getValues());
    shaders.ssao.setInt("width", size.width);
    shaders.ssao.setInt("height", size.height);
    fb.world.bindPositionTexture(0);
    fb.world.bindNormalTexture(1);
    fb.world.bindColorTexture(2);
    fb.world.bindDepthTexture(3);
    ssaoNoise.bind(4);
    fb.ssao.bind();
    rectangle.draw();

    shaders.ssaoBlur.use();
    fb.ssao.bindRedTexture(0);
    fb.ssaoBlur.bind();
    rectangle.draw();
}

void Engine::renderPostWorld() {
    GLWrapper::prepareMainFramebuffer();
    shaders.postWorld.use();
    fb.world.bindColorTexture(0);
    fb.ssaoBlur.bindRedTexture(1);
    fb.world.bindRedTexture(2);
    fb.world.bindNormalTexture(3);
    shaders.postWorld.setInt("ssao", renderSettings.ssao);
    shaders.postWorld.setInt("shadows", renderSettings.shadows);
    rectangle.draw();
}

void Engine::renderTextOverlay(float lag, const Game& game) {
    GLWrapper::disableDepthTesting();
    shaders.text.use();

    Matrix m;
    shaders.text.setMatrix("proj", m.getValues());
    m.translate(-1.0f, 1.0f, 0.0f).scale(2.0f / size.width, -2.0f / size.height, 1.0f);
    shaders.text.setMatrix("view", m.getValues());
    model.clear();
    model.get().scale(2.0f, 2.0f, 2.0f);
    shaders.text.setMatrix("model", model.get().getValues());

    GLWrapper::enableBlending();
    Renderer renderer(shaders.text, model);
    game.renderTextOverlay(lag, renderer, fontRenderer);
    GLWrapper::disableBlending();
}

void Engine::updateWorldProjection() {
    frustum.setProjection(worldProj, size.width, size.height);
    
    if(!renderSettings.shadows) {
        return;
    }

    // http://cgvr.informatik.uni-bremen.de/teaching/cg_literatur/lighthouse3d_view_frustum_culling/index.html
    const float fovY = 60.0f;
    const float nearClip = 0.1f;
    const float farClip = 80.0f;

    float tan = tanf((0.5f * fovY) * M_PI / 180.0f);
    float aspect = (float) size.width / size.height;

    float closeFarClip = 16;
    float nearHigh = tan * nearClip;
    float nearWidth = nearHigh * aspect;
    float farHigh = tan * closeFarClip;
    float farWidth = farHigh * aspect;

    Vector farCenter = camera.getPosition();
    farCenter.addMul(camera.getBack(), -closeFarClip);

    Vector farTopLeft = farCenter;
    farTopLeft.addMul(camera.getRight(), -farWidth).addMul(camera.getUp(), farHigh);

    Vector farBottomLeft = farCenter;
    farBottomLeft.addMul(camera.getRight(), -farWidth).addMul(camera.getUp(), -farHigh);

    Vector farTopRight = farCenter;
    farTopRight.addMul(camera.getRight(), farWidth).addMul(camera.getUp(), farHigh);

    Vector farBottomRight = farCenter;
    farBottomRight.addMul(camera.getRight(), farWidth).addMul(camera.getUp(), -farHigh);

    Vector nearCenter = camera.getPosition();
    nearCenter.addMul(camera.getBack(), -nearClip);

    Vector nearTopLeft = nearCenter;
    nearTopLeft.addMul(camera.getRight(), -nearWidth).addMul(camera.getUp(), nearHigh);

    Vector nearBottomLeft = nearCenter;
    nearBottomLeft.addMul(camera.getRight(), -nearWidth).addMul(camera.getUp(), -nearHigh);

    Vector nearTopRight = nearCenter;
    nearTopRight.addMul(camera.getRight(), nearWidth).addMul(camera.getUp(), nearHigh);

    Vector nearBottomRight = nearCenter;
    nearBottomRight.addMul(camera.getRight(), nearWidth).addMul(camera.getUp(), -nearHigh);

    Vector light(-0.280166, -0.573576, -0.769751);
    Vector lightLeft = light;
    lightLeft.cross(0.0f, 1.0f, 0.0f);
    Vector lightUp = lightLeft;
    lightUp.cross(light);

    Plane plane;
    plane.set(Vector(), light, lightUp);
    float f[8];
    f[0] = plane.getSignedDistance(farTopLeft);
    f[1] = plane.getSignedDistance(farBottomLeft);
    f[2] = plane.getSignedDistance(farTopRight);
    f[3] = plane.getSignedDistance(farBottomRight);
    f[4] = plane.getSignedDistance(nearTopLeft);
    f[5] = plane.getSignedDistance(nearBottomLeft);
    f[6] = plane.getSignedDistance(nearTopRight);
    f[7] = plane.getSignedDistance(nearBottomRight);
    float min = FLT_MAX;
    float max = -FLT_MAX;
    for(uint i = 0; i < 8; i++) {
        if(f[i] < min) {
            min = f[i];
        }
        if(f[i] > max) {
            max = f[i];
        }
    }
    float lightWidth = max - min;

    plane.set(Vector(), light, lightLeft);
    f[0] = plane.getSignedDistance(farTopLeft);
    f[1] = plane.getSignedDistance(farBottomLeft);
    f[2] = plane.getSignedDistance(farTopRight);
    f[3] = plane.getSignedDistance(farBottomRight);
    f[4] = plane.getSignedDistance(nearTopLeft);
    f[5] = plane.getSignedDistance(nearBottomLeft);
    f[6] = plane.getSignedDistance(nearTopRight);
    f[7] = plane.getSignedDistance(nearBottomRight);

    min = FLT_MAX;
    max = -FLT_MAX;
    for(uint i = 0; i < 8; i++) {
        if(f[i] < min) {
            min = f[i];
        }
        if(f[i] > max) {
            max = f[i];
        }
    }
    float lightHeight = max - min;

    // not the real center, but good guess
    renderSettings.testOrthoCenter = nearCenter;
    renderSettings.testOrthoCenter.addMul(camera.getBack(), -closeFarClip * 0.5f);

    if(renderSettings.ortho) {
        worldProj.setToIdentity();
        worldProj.set(0, 2.0f / lightWidth);
        worldProj.set(5, 2.0f / lightHeight);
        worldProj.set(10, -2.0f / (farClip - nearClip));
    }

    worldShadowProj.setToIdentity();
    worldShadowProj.set(0, 2.0f / lightWidth);
    worldShadowProj.set(5, 2.0f / lightHeight);
    worldShadowProj.set(10, -2.0f / (farClip - nearClip));
}

void Engine::updateWorldView() {
    Vector right = camera.getRight();
    Vector up = camera.getUp();
    Vector back = camera.getBack();
    Vector pos = camera.getPosition();

    if(renderSettings.ortho) {
        right.set(0.939693f, 0.0f, -0.34202f);
        back.set(0.280166f, 0.573576f, 0.769751f);
        up.set(-0.196175f, 0.819152f, -0.538986f);
        pos = renderSettings.testOrthoCenter;
    }

    worldView.set(0, right.getX());
    worldView.set(1, up.getX());
    worldView.set(2, back.getX());
    worldView.set(4, right.getY());
    worldView.set(5, up.getY());
    worldView.set(6, back.getY());
    worldView.set(8, right.getZ());
    worldView.set(9, up.getZ());
    worldView.set(10, back.getZ());
    worldView.set(12, right.dotInverse(pos));
    worldView.set(13, up.dotInverse(pos));
    worldView.set(14, back.dotInverse(pos));
    
    if(!renderSettings.shadows) {
        return;
    }

    right.set(0.939693f, 0.0f, -0.34202f);
    back.set(0.280166f, 0.573576f, 0.769751f);
    up.set(-0.196175f, 0.819152f, -0.538986f);
    pos = renderSettings.testOrthoCenter;

    worldShadowView.set(0, right.getX());
    worldShadowView.set(1, up.getX());
    worldShadowView.set(2, back.getX());
    worldShadowView.set(4, right.getY());
    worldShadowView.set(5, up.getY());
    worldShadowView.set(6, back.getY());
    worldShadowView.set(8, right.getZ());
    worldShadowView.set(9, up.getZ());
    worldShadowView.set(10, back.getZ());
    worldShadowView.set(12, right.dotInverse(pos));
    worldShadowView.set(13, up.dotInverse(pos));
    worldShadowView.set(14, back.dotInverse(pos));
}